#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.22 00:00:00                  #
# ================================================== #

__author__ = "Marcin Szczygliński"
__copyright__ = "Copyright 2025, Marcin Szczygliński"
__credits__ = ["Marcin Szczygliński"]
__license__ = "MIT"
__version__ = "2.6.56"
__build__ = "2025-09-22"
__maintainer__ = "Marcin Szczygliński"
__github__ = "https://github.com/szczyglis-dev/py-gpt"
__report__ = "https://github.com/szczyglis-dev/py-gpt/issues"
__website__ = "https://pygpt.net"
__pypi__ = "https://pypi.org/project/pygpt-net"
__snap__ = "https://snapcraft.io/pygpt"
__donate__ = "https://pygpt.net/#donate"
__documentation__ = "https://pygpt.readthedocs.io"
__discord__ = "https://pygpt.net/discord"
__ms_store__ = "https://apps.microsoft.com/detail/xp99r4mx3x65vq"
__email__ = "info@pygpt.net"
__donate_coffee__ = "https://pygpt.net/donate/buymeacoffee"
__donate_paypal__ = "https://pygpt.net/donate/paypal"
__donate_github__ = "https://github.com/sponsors/szczyglis-dev"
