#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# ================================================== #
# This file is a part of PYGPT package               #
# Website: https://pygpt.net                         #
# GitHub:  https://github.com/szczyglis-dev/py-gpt   #
# MIT License                                        #
# Created By  : Marcin Szczygliński                  #
# Updated Date: 2025.09.14 20:00:00                  #
# ================================================== #


class DatabaseDebug:
    def __init__(self, window=None):
        """
        DB debug

        :param window: Window instance
        """
        self.window = window
        self.id = 'db'

    def update(self):
        """Update db window."""
        pass
